# 
#   psprint.tcl
#
#   Description see below.
# 
#     Copyright (C) 2005, Patrick Rammelt <rammelt@cs.tu-berlin.de>
# 
#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with this program; if not, write to the Free Software
#     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
# 
#

#                          ___________________      
# =========================                   ================================
# =========================  Print-Functions  ================================
# =========================___________________================================
#
#

#############################################################################
# NAMESPACE
#############################################################################

namespace eval print {

    # version
    variable v_version "0.1"

    # item-counter
    variable v_itemnr 0
}

#############################################################################
# MAIN FUNCTIONS
#############################################################################

# ---------------------------------------------------------------------------
# return postscript information
# - win      : source window
# - preview  : path for preview window 
#              toplevel window created and destroyed by ::print::ps
# 
# supported widgets:
# - canvas (including embedded windows)
#   embedded windows which contain _scrolled_ canvas widgets are displayed 
#   only when the Img package is available (a scrolled canvas is converted 
#   to an image, which eventually leads to a _much_ bigger postscript file)
# - label
# - entry
# - button
# - menubutton
# - checkbutton
# - radiobutton (displayed the same way as a checkbutton)
# - listbox
#
# notice:
#   for the trick with scrolled canvas widgets it is necessary that the
#   preview canvas has the size of the window which should be displayed ($win)
# ---------------------------------------------------------------------------
proc print::ps {win preview} {
    preview $win $preview
    update
    set ps [$cvs postscript]
    destroy $preview
    return $ps
}


# ---------------------------------------------------------------------------
# opens a dialogue to save postscript file
# (preview window won't be destroyed)
# ---------------------------------------------------------------------------
proc print::print {win preview} {
    preview $win $preview

    $preview.cvs configure -xscrollcommand "$preview.xscroll set" -yscrollcommand "$preview.yscroll set" 
    scrollbar $preview.xscroll -orient horizontal -command "$preview.cvs xview"
    scrollbar $preview.yscroll -orient vertical   -command "$preview.cvs yview"

    frame  $preview.bar -borderwidth 2 -relief sunken
    frame  $preview.bar.space1
    button $preview.bar.print -text "Save PostScript" -command "::print::save $win $preview"
    frame  $preview.bar.space2
    button $preview.bar.close -text "Close" -command "destroy $preview"

    grid $preview.bar.space1 -column 2 -row 0
    grid $preview.bar.print  -column 1 -row 0
    grid $preview.bar.space2 -column 2 -row 0
    grid $preview.bar.close  -column 3 -row 0
    grid columnconfigure $preview.bar 2 -weight 1
    grid columnconfigure $preview.bar 0 -weight 1

    grid $preview.bar     -column 0 -row 2 -sticky ew -columnspan 2
    grid $preview.xscroll -column 0 -row 1 -sticky ew
    grid $preview.yscroll -column 1 -row 0 -sticky ns
}


# ---------------------------------------------------------------------------
# creates the preview window
# ---------------------------------------------------------------------------
proc print::preview {win preview} {
    if {[winfo exists $preview]} { bell; return }
    if {![winfo exists $win]}    { bell; return }

    if {[catch { set bg [$win cget -background] }]} { set bg grey }

    set w [winfo width  $win]
    set h [winfo height $win]

    set cvs $preview.cvs.cvs

    toplevel $preview
    canvas $preview.cvs -background grey \
	-scrollregion [list -5 -5 [expr $w+5] [expr $h+5]] -width [expr $w+10] -height [expr $h+10]
    canvas $cvs -background $bg -width $w -height $h -scrollregion [list 0 0 [expr $w-1] [expr $h-1]] 
    $preview.cvs create window 0 0 -window $cvs -anchor nw

    grid $preview.cvs -column 0 -row 0 -sticky nsew
    grid columnconfigure $preview 0 -weight 1
    grid rowconfigure    $preview 0 -weight 1

    viewWindow $win $cvs
    bind $cvs <Destroy> "::print::unset_images $cvs"
}


#############################################################################
# CALLBACK
#############################################################################

# ---------------------------------------------------------------------------
# save
# ---------------------------------------------------------------------------
proc print::save {cvs preview} {
    set cvs $preview.cvs.cvs
    set types {{"Postscript" ".eps"} {"All Files" "*"}}
    set file [tk_getSaveFile -title "Save Postscript" -filetypes $types -defaultextension .eps]
    if {[llength $file]} {
	set ps [$cvs postscript]
	if {![catch {set outputfile [open "$file" "w"]}]!=0} {
	    puts $outputfile $ps
	    close $outputfile
	    destroy $preview
	} else { bell }
    }
}


#############################################################################
# PARSE WINDOWS AND FILL PREVIEW
#############################################################################

# ---------------------------------------------------------------------------
# transform widgets (win + children of win) to canvas items in pre
# ---------------------------------------------------------------------------
proc print::viewWindow {win pre {offset {0 0}} {rec 0}} {
    # recursion counter
    incr rec 1

    # check
    if {![winfo ismapped $win]} { return }

    foreach {x0 y0} $offset {}
    set dx 0; set dy 0
    #catch { set dx [winfo pixels $win [$win cget -padx]] }
    #catch { set dy [winfo pixels $win [$win cget -pady]] }
    set x  [expr $x0 + $dx]
    set y  [expr $y0 + $dy]
    set w  [expr [winfo width  $win] - (2*$dx)]
    set h  [expr [winfo height $win] - (2*$dy)]
    set b  [$win cget -borderwidth]
    set bg [$win cget -background]

    # "parse" widgets
    switch -glob -- [winfo class $win] {
	Canvas* { 
	    createCanvas $pre $x $y $w $h $bg $bg $b [$win cget -relief] $win $rec
	}
	Frame* { 
	    set item [createBox $pre $x $y $w $h $bg $bg $b [$win cget -relief]]
	    viewSubWindows $win $pre $offset $rec
	}
	Label* { 
	    createLabel $pre $x $y $w $h $bg $bg $b [$win cget -relief] [$win cget -text] \
		[$win cget -font] [$win cget -foreground] [$win cget -justify] [$win cget -image] }
	Entry* { 
	    createLabel $pre $x $y $w $h $bg $bg $b [$win cget -relief] [set [$win cget -textvariable]] \
		[$win cget -font] [$win cget -foreground] [$win cget -justify] {} }
	Button* { 
	    createLabel $pre $x $y $w $h $bg $bg $b [$win cget -relief] [$win cget -text] \
		[$win cget -font] [$win cget -foreground] [$win cget -justify] [$win cget -image] }
	Menubutton*  { 
	    createLabel $pre $x $y $w $h $bg $bg $b [$win cget -relief] [$win cget -text] \
		[$win cget -font] [$win cget -foreground] [$win cget -justify] [$win cget -image] }
	Scrollbar* { 
	    createScrollBar $pre $x $y $w $h $bg $bg $b [$win cget -relief] [$win get] [$win cget -orient]
	}
	Listbox* { 
	    createListBox $pre $x $y $w $h $bg $bg $b [$win cget -relief] $win
	}
	Checkbutton* { 
	    createCheckButton $pre $x $y $w $h $bg $bg $b [$win cget -relief] [$win cget -text] \
		[$win cget -font] [$win cget -foreground] [$win cget -image] \
		[set [$win cget -variable]] [$win cget -onvalue] \
		[$win cget -selectcolor] [$win cget -selectimage] 
	}
	Radiobutton* { 
	    createCheckButton $pre $x $y $w $h $bg $bg $b [$win cget -relief] [$win cget -text] \
		[$win cget -font] [$win cget -foreground] [$win cget -image] \
		[set [$win cget -variable]] [$win cget -value] \
		[$win cget -selectcolor] [$win cget -selectimage] 
	} 
	Text* {
	    createBox $pre $x $y $w $h $bg $bg $b [$win cget -relief]
	    $pre create text [expr $x+int($w/2)] [expr $y+int($h/2)] -anchor c \
		    -text "TEXT WIDGET\n(not supported)" -fill red -justify center
	}
	Toplevel* { set rec 0 }
	default {
	    createBox $pre $x $y $w $h $bg $bg $b flat
	    $pre create text [expr $x+int($w/2)] [expr $y+int($h/2)] -anchor c \
		    -text "[winfo class $win]\n(not supported)" -fill red -justify center
	}
    }
}

	    
# ---------------------------------------------------------------------------
# recursively preview subwindows
# ---------------------------------------------------------------------------
proc print::viewSubWindows {win pre {offset {0 0}} {rec 0}} {
    foreach {x0 y0} $offset {}
    set x1 [winfo rootx  $win]
    set y1 [winfo rooty  $win]
    foreach child [winfo children $win] { 
	switch -- [winfo manager $child] {
	    grid    { set ok 1 }
	    pack    { set ok 1 }
	    place   { set ok 1 }
	    wm      { set ok 0 }
	    default { set ok 0 }
	}
	if {$ok} {
	    set xd [expr $x0 + [winfo rootx $child] - $x1]
	    set yd [expr $y0 + [winfo rooty $child] - $y1]
	    viewWindow $child $pre [list $xd $yd] $rec
	}
    }
}


# ---------------------------------------------------------------------------
# copy canvas items from cvs to pre (including embedded windows)
# ---------------------------------------------------------------------------
proc print::viewCanvas {cvs pre {offset {0 0}} {rec 0}} {
    foreach {x0 y0} $offset {}

    # sort
    set windows {}
    set items {}
    foreach item [$cvs find all] {
	set itype  [$cvs type   $item]
	switch -- $itype {
	    window  { lappend windows $item }
	    image   { lappend items   $item }
	    default { lappend items   $item }
	}
    }

    # scroll offsets
    set xd [$cvs canvasx 0]
    set yd [$cvs canvasy 0]

    # items first
    foreach item $items {
	set itype  [$cvs type   $item]
	set coords [$cvs coords $item]
	set coords2 {}
	foreach {x y} $coords { lappend coords2 [expr $x+$x0 - $xd] [expr $y+$y0 - $yd] }
	eval $pre create $itype $coords2 [item_config $cvs $item]
    }

    # windows
    foreach item $windows {
	foreach {x y} [$cvs coords $item] {}
	set win [$cvs itemcget $item -window]
	set x [expr $x+$x0 - $xd]
	set y [expr $y+$y0 - $yd]
	set w [winfo width  $win]
	set h [winfo height $win]
	switch -glob -- [$cvs itemcget $item -anchor] {
	    so { set x [expr $x - $w + 1]; set y [expr $y - $h + 1] }
	    sw { set y [expr $y - $h + 1] }
	    no { set x [expr $x - $w + 1] }
	    c* { 
		set x [expr $x - int($w/2) + int($w)%2] 
		set y [expr $y - int($h/2) + int($h)%2] 
	    }
	}
	viewWindow $win $pre [list $x $y] $rec
    }
}


#############################################################################
# SCROLLED CANVAS - SPECIAL TREATMENT
#############################################################################

# ---------------------------------------------------------------------------
# create image of canvas
# ---------------------------------------------------------------------------
proc print::canvas2image {cvs pre} {
    variable v_img
    if {![catch {package require Img}]} {
	set img [image create photo -format window -data $cvs]
	lappend v_img($pre) $img
	return $img
    }
    return ""
}


# ---------------------------------------------------------------------------
# delete images
# ---------------------------------------------------------------------------
proc print::unset_images {pre} {
    variable v_img
    foreach {elem images} [array get v_img $pre] { 
	foreach img $images { image delete $img }
    }
    array unset v_img $pre
}


#############################################################################
# CREATE PREVIEW-ITEMS
#############################################################################

# ---------------------------------------------------------------------------
# create rectangle with 3d-shape
# ---------------------------------------------------------------------------
proc print::createBox {cvs x1 y1 w h shapecolor innercolor width relief} {
    set tag [newtag]

    set width [max 0 [expr $width - (1-($width%2))]]

    set x2 [expr $x1 + $w - 1]
    set y2 [expr $y1 + $h - 1]
    set col3d [color3d $cvs $shapecolor]

    switch -- $relief {
	sunken {
	    set color1 [lindex $col3d 0]
	    set color2 [lindex $col3d 1]
	}
	raised {
	    set color1 [lindex $col3d 1]
	    set color2 [lindex $col3d 0]
	}
	default { 
	    set color1 $shapecolor
	    set color2 $shapecolor
	}

    }

    set linewidth 1
    $cvs create rectangle [expr $x1 + $width] [expr $y1 + $width] \
	[expr $x2 - $width] [expr $y2 - $width] -fill $innercolor \
	-outline $innercolor -width $linewidth -tags $tag
    if {$width>0} {
	$cvs create polygon $x1 $y1 $x2 $y1 \
	    [expr $x2 - $width + 1] [expr $y1 + $width - 1] \
	    [expr $x1 + $width - 1] [expr $y1 + $width - 1] -fill $color1 \
	    -outline $color1 -width $linewidth -tags $tag
	$cvs create polygon $x1 $y2 $x2 $y2 \
	    [expr $x2 - $width + 1] [expr $y2 - $width + 1] \
	    [expr $x1 + $width - 1] [expr $y2 - $width + 1] -fill $color2 \
	    -outline $color2 -width $linewidth -tags $tag
	$cvs create polygon $x1 $y1 $x1 $y2 \
	    [expr $x1 + $width - 1] [expr $y2 - $width + 1] \
	    [expr $x1 + $width - 1] [expr $y1 + $width - 1] -fill $color1 \
	    -outline $color1 -width $linewidth -tags $tag
	$cvs create polygon $x2 $y1 $x2 $y2 \
	    [expr $x2 - $width + 1] [expr $y2 - $width + 1] \
	    [expr $x2 - $width + 1] [expr $y1 + $width - 1] -fill $color2 \
	    -outline $color2 -width $linewidth -tags $tag
    }

    return $tag
}


# ---------------------------------------------------------------------------
# create rectangle with 3d-shape + text or image
# ---------------------------------------------------------------------------
proc print::createLabel {cvs x y w h shapecolor innercolor width relief text font fg justify image} {
    set tag [newtag]
    set item [createBox $cvs $x $y $w $h $shapecolor $innercolor $width $relief]
    $cvs addtag $tag withtag $item

    if {[llength $image]} {
	set item [$cvs create image [expr $x+int($w/2)] [expr $y+int($h/2)] -image $image -anchor c]
	fit $cvs $item $x $y $w $h
    } elseif {[llength $text]}  {
	switch -glob -- $justify {
	    l*      { set item [$cvs create text [expr $x+$width+2] [expr $y+int($h/2)] -anchor w \
				    -justify $justify -text $text -fill $fg -font $font] }
	    r*      { set item [$cvs create text [expr $x+$w-1-$width-2] [expr $y+int($h/2)] -anchor e \
				    -justify $justify -text $text -fill $fg -font $font] }
	    default { set item [$cvs create text [expr $x+int($w/2)] [expr $y+int($h/2)] -anchor c \
				    -justify $justify -text $text -fill $fg -font $font] }
	}
	fit $cvs $item $x $y $w $h
    }

    return $tag
}


# ---------------------------------------------------------------------------
# create canvas with all its contents including subwindows
# ---------------------------------------------------------------------------
proc print::createCanvas {cvs x y w h shapecolor innercolor width relief win rec} {
    set tag [newtag]
    set item [createBox $cvs $x $y $w $h $shapecolor $innercolor $width $relief]
    $cvs addtag $tag withtag $item

    if {$rec==1 || ![isscrolled $win]} { 
	viewCanvas $win $cvs [list [expr $x + $width] [expr $y + $width]] $rec
    } else {
	set xx [expr $x + $width]
	set yy [expr $y + $width]
	set ww [expr $w - 2 * $width]
	set hh [expr $h - 2 * $width]
	set sub $cvs.sub
	canvas $sub -width $ww -height $hh -background $innercolor -borderwidth 0 -highlightthickness 0
	$cvs create window $xx $yy -window $sub -tags $sub -anchor nw
	viewCanvas $win $sub [list 0 0] $rec
	update
	if {[winfo ismapped $sub]} {
	    if {[llength [set img [canvas2image $sub $cvs]]]} {
		$cvs create image $xx $yy -image $img -anchor nw -tags $tag
	    } else {
		$cvs create text [expr $x+int($w/2)] [expr $y+int($h/2)] -anchor c -fill red \
			    -text "SCROLLED CANVAS\n(need Img-package)" -justify center -tags $tag
	    }
	}
	$cvs delete $sub
	destroy $sub
    }
}


# ---------------------------------------------------------------------------
# create canvas scrollbar represantation
# ---------------------------------------------------------------------------
proc print::createScrollBar {cvs x y w h shapecolor innercolor width relief range orientation} {
    set tag [newtag]
    set item [createBox $cvs $x $y $w $h $shapecolor $innercolor $width sunken]
    $cvs addtag $tag withtag $item

    set x1 [expr $x + $width]
    set y1 [expr $y + $width]
    set w1 [expr $w - (2 * $width)]
    set h1 [expr $h - (2 * $width)]
    set x2 [expr $x1 + $w1 - 1]
    set y2 [expr $y1 + $h1 - 1]
    foreach {first last} $range {}
    switch -glob -- $orientation {
	hor* { 
	    set a [expr $h - (2 * $width)]
	    set x3 [expr $x1 + $a]
	    set w3 [expr $w1 - (2 * $a)]
	    set item1 [createBox $cvs $x1 $y1 $a $a $shapecolor $innercolor 2 raised]
	    set item2 [createBox $cvs [expr $x2-$a+1] $y1 $a $a $shapecolor $innercolor 2 raised]
	    set item3 [createBox $cvs [expr $x3 + int($first * $w3)] $y1 [expr int($w3 * ($last-$first))] $h1 \
			   $shapecolor $innercolor 2 raised]
	}
	ver* { 
	    set a [expr $w - (2 * $width)]
	    set y3 [expr $y1 + $a]
	    set h3 [expr $h1 - (2 * $a)]
	    set item1 [createBox $cvs $x1 $y1 $a $a $shapecolor $innercolor 2 raised]
	    set item2 [createBox $cvs $x1 [expr $y2-$a+1] $a $a $shapecolor $innercolor 2 raised]
	    set item3 [createBox $cvs $x1 [expr $y3 + int($first * $h3)] $w1 [expr int($h3 * ($last-$first))] \
			   $shapecolor $innercolor 2 raised]
	}
    }
    $cvs addtag $tag withtag $item1
    $cvs addtag $tag withtag $item2
    $cvs addtag $tag withtag $item3
}


# ---------------------------------------------------------------------------
# create canvas listbox represantation
# ---------------------------------------------------------------------------
proc print::createListBox {cvs x y w h shapecolor innercolor width relief win} {
    set tag [newtag]
    set item [createBox $cvs $x $y $w $h $shapecolor $innercolor $width $relief]
    $cvs addtag $tag withtag $item

    set fg   [$win cget -foreground]
    set font [$win cget -font]
    for {set i 0} {$i<[$win size]} {incr i} {
	foreach {x1 y1 x2 y2} [$win bbox $i] {
	    set item [$cvs create text [expr $x+$x1] [expr $y+$y1] -anchor nw -tags $tag \
			  -text [lindex [$win get $i $i] 0] -fill $fg -font $font] 
	    fit $cvs $item $x $y $w $h
	}
    }
}


# ---------------------------------------------------------------------------
# create canvas checkbutton represantation
# ---------------------------------------------------------------------------
proc print::createCheckButton {cvs x y w h shapecolor innercolor width relief text \
				 font fg image value onvalue selectcolor selectimage} {
    set tag [newtag]
    set item [createBox $cvs $x $y $w $h $shapecolor $innercolor $width $relief]
    $cvs addtag $tag withtag $item

    if {[llength $selectimage]} { 
	set wsel [image width  $selectimage]
	set hsel [image height $selectimage] 
    } else { 
	set wsel 10
	set hsel 10 
    }

    if {![string compare $value $onvalue]} { 
	set item1 [createLabel $cvs [expr $x+2] [expr $y+($h-$hsel)/2] $wsel $hsel \
		       $shapecolor $selectcolor 2 sunken {} {} {} {} $selectimage]
    } else { 
	set item1 [createBox $cvs [expr $x+2] [expr $y+($h-10)/2] 10 10 \
		       $shapecolor $innercolor 2 raised]
    }
    if {[llength $image]} {
	set item2 [$cvs create image [expr $x+$wsel+5] [expr $y+int($h/2)] \
		       -image $image -anchor w]
    } else {
	set item2 [$cvs create text [expr $x+$wsel+5] [expr $y+int($h/2)] \
		       -anchor w -text $text -fill $fg -font $font]
    }
    $cvs addtag $tag withtag $item1
    $cvs addtag $tag withtag $item2
    fit $cvs $item1 $x $y $w $h
    fit $cvs $item2 $x $y $w $h
}


#############################################################################
# HELPERS
#############################################################################

# ---------------------------------------------------------------------------
# get tag for new item
# ---------------------------------------------------------------------------
proc print::newtag {} {
    variable v_itemnr
    return [format "item%s" [incr v_itemnr]]
}


# ---------------------------------------------------------------------------
# check size and delete item if it does not fit
# ---------------------------------------------------------------------------
proc print::fit {cvs item x y w h} {
    foreach {x1 y1 x2 y2} [$cvs bbox $item] {}
    if {[expr $x1<$x || $y1<$y || $x2>($x+$w) || $y2>($y+$h)]} { 
	$cvs delete $item 
    }
}


# ---------------------------------------------------------------------------
# get item options
# ---------------------------------------------------------------------------
proc print::item_config {cvs item} {
    set config {}
    foreach opt [$cvs itemconfigure $item] { 
	foreach {o a b c d} $opt { lappend config $o [$cvs itemcget $item $o] }
    }
    return $config
}


# ---------------------------------------------------------------------------
# get colors for 3d-shape
# ---------------------------------------------------------------------------
proc print::color3d {path bgcolor} {
    switch -- $bgcolor {
	"" { return [list "" ""] }
	default {
	    foreach val [winfo rgb $path $bgcolor] {
		lappend dark [expr 60*$val/100]
		set tmp1 [expr 14*$val/10]
		if {$tmp1 > 65535} { set tmp1 65535 }
		set tmp2 [expr (65535+$val)/2]
		lappend light [expr ($tmp1 > $tmp2) ? $tmp1 : $tmp2]
	    }
	    return [list [eval format "#%04x%04x%04x" $dark] [eval format "#%04x%04x%04x" $light]]
	}
    }
}


# ---------------------------------------------------------------------------
# min, max
# ---------------------------------------------------------------------------
proc print::min {a b} { return [expr ($a<$b ? $a : $b)] }
proc print::max {a b} { return [expr ($a>$b ? $a : $b)] }


# ---------------------------------------------------------------------------
# scrolled widget (canvas) ?
# ---------------------------------------------------------------------------
proc print::isscrolled {win} {
    foreach {xstart xend} [$win xview] {}
    foreach {ystart yend} [$win yview] {}
    return [expr $xstart>0.0 || $xend<1.0 || $ystart>0.0 || $yend<1.0]
}

